@@ -27,6 +27,7 @@
 #include <locale.h>
 #include <string.h>
 #include <strings.h>
+#include <dbus/dbus.h>
 #include "ibusimpl.h"
 #include "dbusimpl.h"
 #include "server.h"
@@ -56,6 +57,13 @@ static void     bus_ibus_impl_add_factory       (BusIBusImpl        *ibus,
                                                  BusFactoryProxy    *factory);
 static void     bus_ibus_impl_set_trigger       (BusIBusImpl        *ibus,
                                                  GValue             *value);
+static void     bus_ibus_impl_set_next_engine_in_menu
+                                                (BusIBusImpl        *ibus,
+                                                 GValue             *value);
+static void     bus_ibus_impl_set_previous_engine
+                                                (BusIBusImpl        *ibus,
+                                                 GValue             *value);
+
 static void     bus_ibus_impl_set_preload_engines
                                                 (BusIBusImpl        *ibus,
                                                  GValue             *value);
@@ -65,6 +73,9 @@ static void     bus_ibus_impl_set_use_sys_layout
 static void     bus_ibus_impl_set_embed_preedit_text
                                                 (BusIBusImpl        *ibus,
                                                  GValue             *value);
+static void     bus_ibus_impl_set_enable_by_default
+                                                (BusIBusImpl        *ibus,
+                                                 GValue             *value);
 
 static void     bus_ibus_impl_set_use_global_engine
                                                 (BusIBusImpl        *ibus,
@@ -73,9 +84,26 @@ static void     bus_ibus_impl_set_global_engine (BusIBusImpl        *ibus,
                                                  BusEngineProxy     *engine);
 
 static void     bus_ibus_impl_registry_changed  (BusIBusImpl        *ibus);
+static void     bus_ibus_impl_global_engine_changed
+                                                (BusIBusImpl        *ibus);
+
 static void     _factory_destroy_cb             (BusFactoryProxy    *factory,
                                                  BusIBusImpl        *ibus);
 
+static void     bus_ibus_impl_set_context_engine(BusIBusImpl        *ibus,
+                                                 BusInputContext    *context,
+                                                 BusEngineProxy     *engine);
+
+static gchar   *bus_ibus_impl_load_global_engine_name_from_config
+                                                (BusIBusImpl        *ibus);
+static void     bus_ibus_impl_save_global_engine_name_to_config
+                                                (BusIBusImpl        *ibus);
+
+static gchar   *bus_ibus_impl_load_global_previous_engine_name_from_config
+                                                (BusIBusImpl        *ibus);
+static void     bus_ibus_impl_save_global_previous_engine_name_to_config
+                                                (BusIBusImpl        *ibus);
+
 G_DEFINE_TYPE(BusIBusImpl, bus_ibus_impl, IBUS_TYPE_SERVICE)
 
 BusIBusImpl *
@@ -164,18 +192,18 @@ bus_ibus_impl_set_trigger (BusIBusImpl *ibus,
 }
 
 static void
-bus_ibus_impl_set_next_engine (BusIBusImpl *ibus,
-                               GValue      *value)
+bus_ibus_impl_set_next_engine_in_menu (BusIBusImpl *ibus,
+                                       GValue      *value)
 {
-    GQuark hotkey = g_quark_from_static_string ("next-engine");
+    GQuark hotkey = g_quark_from_static_string ("next-engine-in-menu");
     bus_ibus_impl_set_hotkey (ibus, hotkey, value);
 }
 
 static void
-bus_ibus_impl_set_prev_engine (BusIBusImpl *ibus,
-                               GValue      *value)
+bus_ibus_impl_set_previous_engine (BusIBusImpl *ibus,
+                                   GValue      *value)
 {
-    GQuark hotkey = g_quark_from_static_string ("prev-engine");
+    GQuark hotkey = g_quark_from_static_string ("previous-engine");
     bus_ibus_impl_set_hotkey (ibus, hotkey, value);
 }
 
@@ -234,12 +262,21 @@ bus_ibus_impl_set_use_sys_layout (BusIBusImpl *ibus,
 }
 
 static void
-bus_ibus_impl_set_embed_preedit_text (BusIBusImpl        *ibus,
-							     GValue             *value){
+bus_ibus_impl_set_embed_preedit_text (BusIBusImpl *ibus,
+                                      GValue      *value)
+{
     if (value != NULL && G_VALUE_TYPE (value) == G_TYPE_BOOLEAN) {
         ibus->embed_preedit_text = g_value_get_boolean (value);
     }
+}
 
+static void
+bus_ibus_impl_set_enable_by_default (BusIBusImpl *ibus,
+                                     GValue      *value)
+{
+    if (value != NULL && G_VALUE_TYPE (value) == G_TYPE_BOOLEAN) {
+        ibus->enable_by_default = g_value_get_boolean (value);
+    }
 }
 
 static void
@@ -271,6 +308,8 @@ bus_ibus_impl_set_use_global_engine (BusIBusImpl *ibus,
         /* turn off use_global_engine option */
         bus_ibus_impl_set_global_engine (ibus, NULL);
         ibus->use_global_engine = FALSE;
+
+        g_free (ibus->global_previous_engine_name);
     }
 }
 
@@ -356,12 +395,17 @@ bus_ibus_impl_reload_config (BusIBusImpl *ibus)
         void ( *func) (BusIBusImpl *, GValue *);
     } entries [] = {
         { "general/hotkey", "trigger", bus_ibus_impl_set_trigger },
-        { "general/hotkey", "next_engine", bus_ibus_impl_set_next_engine },
-        { "general/hotkey", "prev_engine", bus_ibus_impl_set_prev_engine },
+        /* Only for backward compatibility, shall be removed later. */
+        { "general/hotkey", "next_engine", bus_ibus_impl_set_next_engine_in_menu },
+        { "general/hotkey", "next_engine_in_menu", bus_ibus_impl_set_next_engine_in_menu },
+        /* Only for backward compatibility, shall be removed later. */
+        { "general/hotkey", "prev_engine", bus_ibus_impl_set_previous_engine },
+        { "general/hotkey", "previous_engine", bus_ibus_impl_set_previous_engine },
         { "general", "preload_engines", bus_ibus_impl_set_preload_engines },
         { "general", "use_system_keyboard_layout", bus_ibus_impl_set_use_sys_layout },
         { "general", "use_global_engine", bus_ibus_impl_set_use_global_engine },
         { "general", "embed_preedit_text", bus_ibus_impl_set_embed_preedit_text },
+        { "general", "enable_by_default", bus_ibus_impl_set_enable_by_default },
     };
 
     for (i = 0; i < G_N_ELEMENTS (entries); i++) {
@@ -399,13 +443,18 @@ _config_value_changed_cb (IBusConfig  *config,
         gchar *key;
         void ( *func) (BusIBusImpl *, GValue *);
     } entries [] = {
-        { "general/hotkey", "trigger",     bus_ibus_impl_set_trigger },
-        { "general/hotkey", "next_engine", bus_ibus_impl_set_next_engine },
-        { "general/hotkey", "prev_engine", bus_ibus_impl_set_prev_engine },
+        { "general/hotkey", "trigger", bus_ibus_impl_set_trigger },
+        /* Only for backward compatibility, shall be removed later. */
+        { "general/hotkey", "next_engine", bus_ibus_impl_set_next_engine_in_menu },
+        { "general/hotkey", "next_engine_in_menu", bus_ibus_impl_set_next_engine_in_menu },
+        /* Only for backward compatibility, shall be removed later. */
+        { "general/hotkey", "prev_engine", bus_ibus_impl_set_previous_engine },
+        { "general/hotkey", "previous_engine", bus_ibus_impl_set_previous_engine },
         { "general", "preload_engines",    bus_ibus_impl_set_preload_engines },
         { "general", "use_system_keyboard_layout", bus_ibus_impl_set_use_sys_layout },
         { "general", "use_global_engine", bus_ibus_impl_set_use_global_engine },
         { "general", "embed_preedit_text", bus_ibus_impl_set_embed_preedit_text },
+        { "general", "enable_by_default", bus_ibus_impl_set_enable_by_default },
     };
 
     for (i = 0; i < G_N_ELEMENTS (entries); i++) {
@@ -554,8 +603,10 @@ bus_ibus_impl_init (BusIBusImpl *ibus)
 
     ibus->use_sys_layout = FALSE;
     ibus->embed_preedit_text = TRUE;
+    ibus->enable_by_default = FALSE;
     ibus->use_global_engine = FALSE;
     ibus->global_engine = NULL;
+    ibus->global_previous_engine_name = NULL;
 
     bus_ibus_impl_reload_config (ibus);
 
@@ -631,6 +682,8 @@ bus_ibus_impl_destroy (BusIBusImpl *ibus)
         ibus->global_engine = NULL;
     }
 
+    g_free (ibus->global_previous_engine_name);
+
     bus_server_quit (BUS_DEFAULT_SERVER);
     ibus_object_destroy ((IBusObject *) BUS_DEFAULT_SERVER);
     IBUS_OBJECT_CLASS(bus_ibus_impl_parent_class)->destroy (IBUS_OBJECT (ibus));
@@ -677,8 +730,25 @@ _ibus_introspect (BusIBusImpl     *ibus,
         "      <arg name=\"data\" direction=\"in\" type=\"v\"/>\n"
         "      <arg name=\"data\" direction=\"out\" type=\"v\"/>\n"
         "    </method>\n"
+        "    <method name=\"GetUseSysLayout\">\n"
+        "      <arg name=\"enable\" direction=\"out\" type=\"b\"/>\n"
+        "    </method>\n"
+        "    <method name=\"GetUseGlobalEngine\">\n"
+        "      <arg name=\"enable\" direction=\"out\" type=\"b\"/>\n"
+        "    </method>\n"
+        "    <method name=\"GetGlobalEngine\">\n"
+        "      <arg name=\"desc\" direction=\"out\" type=\"v\"/>\n"
+        "    </method>\n"
+        "    <method name=\"SetGlobalEngine\">\n"
+        "      <arg name=\"name\" direction=\"in\" type=\"s\"/>\n"
+        "    </method>\n"
+        "    <method name=\"IsGlobalEngineEnabled\">\n"
+        "      <arg name=\"enable\" direction=\"out\" type=\"b\"/>\n"
+        "    </method>\n"
         "    <signal name=\"RegistryChanged\">\n"
         "    </signal>\n"
+        "    <signal name=\"GlobalEngineChanged\">\n"
+        "    </signal>\n"
         "  </interface>\n"
         "</node>\n";
 
@@ -794,13 +864,29 @@ _context_request_engine_cb (BusInputContext *context,
     g_return_if_fail (bus_input_context_has_focus (context));
 
     if (engine_name == NULL || engine_name[0] == '\0') {
+        /* Use global engine if possible. */
+        if (ibus->use_global_engine) {
+            if (ibus->global_engine) {
+                bus_ibus_impl_set_context_engine (ibus, context, ibus->global_engine);
+                return;
+            }
+            else {
+                engine_name = bus_ibus_impl_load_global_engine_name_from_config (ibus);
+                if (engine_name) {
+                    engine_desc = _find_engine_desc_by_name (ibus, engine_name);
+                    g_free (engine_name);
+                }
+            }
+        }
         /* request default engine */
-        if (ibus->register_engine_list) {
-            engine_desc = (IBusEngineDesc *)ibus->register_engine_list->data;
-        }
-        else if (ibus->engine_list) {
-            engine_desc = (IBusEngineDesc *)ibus->engine_list->data;
-        }
+        if (!engine_desc) {
+            if (ibus->register_engine_list) {
+                engine_desc = (IBusEngineDesc *)ibus->register_engine_list->data;
+            }
+            else if (ibus->engine_list) {
+                engine_desc = (IBusEngineDesc *)ibus->engine_list->data;
+            }
+         }
     }
     else {
         /* request engine by name */
@@ -810,14 +896,14 @@ _context_request_engine_cb (BusInputContext *context,
     if (engine_desc != NULL) {
         engine = bus_ibus_impl_create_engine (engine_desc);
         if (engine != NULL) {
-            bus_input_context_set_engine (context, engine);
+            bus_ibus_impl_set_context_engine (ibus, context, engine);
         }
     }
 }
 
 static void
-_context_request_next_engine_cb (BusInputContext *context,
-                                 BusIBusImpl     *ibus)
+bus_ibus_impl_context_request_next_engine_in_menu (BusIBusImpl     *ibus,
+                                                   BusInputContext *context)
 {
     BusEngineProxy *engine;
     IBusEngineDesc *desc;
@@ -858,16 +944,27 @@ _context_request_next_engine_cb (BusInputContext *context,
     if (next_desc != NULL) {
         engine = bus_ibus_impl_create_engine (next_desc);
         if (engine != NULL) {
-            bus_input_context_set_engine (context, engine);
+            bus_ibus_impl_set_context_engine (ibus, context, engine);
         }
     }
 }
 
 static void
-_context_request_prev_engine_cb (BusInputContext *context,
-                                 BusIBusImpl     *ibus)
+bus_ibus_impl_context_request_previous_engine (BusIBusImpl     *ibus,
+                                               BusInputContext *context)
 {
+    gchar *engine_name = NULL;
+    if (!ibus->use_global_engine) {
+        engine_name = (gchar *) g_object_get_data (G_OBJECT (context), "previous-engine-name");
+    }
+    else {
+        if (!ibus->global_previous_engine_name) {
+            ibus->global_previous_engine_name = bus_ibus_impl_load_global_previous_engine_name_from_config (ibus);
+        }
+        engine_name = ibus->global_previous_engine_name;
+    }
 
+    _context_request_engine_cb (context, engine_name, ibus);
 }
 
 static void
@@ -894,7 +991,12 @@ bus_ibus_impl_set_global_engine (BusIBusImpl    *ibus,
         return;
     }
 
+    g_free (ibus->global_previous_engine_name);
+    ibus->global_previous_engine_name = NULL;
     if (ibus->global_engine) {
+        /* Save the current global engine's name as previous engine. */
+        ibus->global_previous_engine_name = g_strdup (bus_engine_proxy_get_desc (ibus->global_engine)->name);
+
         ibus_object_destroy ((IBusObject *)ibus->global_engine);
         /* global_engine should be NULL */
         g_assert (ibus->global_engine == NULL);
@@ -906,6 +1008,30 @@ bus_ibus_impl_set_global_engine (BusIBusImpl    *ibus,
         g_signal_connect (ibus->global_engine, "destroy",
                 G_CALLBACK (_global_engine_destroy_cb), ibus);
     }
+
+    bus_ibus_impl_save_global_engine_name_to_config (ibus);
+    bus_ibus_impl_save_global_previous_engine_name_to_config (ibus);
+    bus_ibus_impl_global_engine_changed (ibus);
+}
+
+static void
+bus_ibus_impl_set_context_engine (BusIBusImpl     *ibus,
+                                  BusInputContext *context,
+                                  BusEngineProxy  *engine) {
+  g_object_set_data (G_OBJECT (context), "previous-engine-name", NULL);
+
+  /* If use_global_engine is disabled, then we need to save the previous engine
+   * of each input context. */
+  if (!ibus->use_global_engine) {
+      BusEngineProxy *previous_engine = bus_input_context_get_engine (context);
+      if (previous_engine) {
+          g_object_set_data_full (G_OBJECT (context), "previous-engine-name",
+                                  g_strdup (bus_engine_proxy_get_desc (previous_engine)->name),
+                                  g_free);
+      }
+  }
+
+  bus_input_context_set_engine (context, engine);
 }
 
 static void
@@ -914,8 +1040,7 @@ _context_engine_changed_cb (BusInputContext *context,
 {
     BusEngineProxy *engine;
 
-    if (context != ibus->focused_context ||
-        ibus->use_global_engine != TRUE) {
+    if (context != ibus->focused_context || !ibus->use_global_engine) {
         return;
     }
 
@@ -953,7 +1078,7 @@ _context_focus_out_cb (BusInputContext    *context,
     /* If the use_global_engine option is enabled,
      * the global engine shoule be detached from the focused context. */
     if (ibus->use_global_engine) {
-        bus_input_context_set_engine (context, NULL);
+        bus_ibus_impl_set_context_engine (ibus, context, NULL);
     }
 
     g_object_unref (context);
@@ -989,10 +1114,14 @@ _context_focus_in_cb (BusInputContext *context,
      * - Set the context's enabled state according to the saved state.
      * Note: we get this signal only if the context supports IBUS_CAP_FOCUS. */
     if (ibus->use_global_engine) {
-        bus_input_context_set_engine (context, ibus->global_engine);
-        if (ibus->global_engine &&
-            bus_engine_proxy_is_enabled (ibus->global_engine)) {
-            bus_input_context_enable (context);
+        if (!ibus->global_engine) {
+            bus_ibus_impl_set_global_engine (ibus, bus_input_context_get_engine (context));
+        }
+        else {
+            bus_ibus_impl_set_context_engine (ibus, context, ibus->global_engine);
+            if (ibus->global_engine && bus_engine_proxy_is_enabled (ibus->global_engine)) {
+                bus_input_context_enable (context);
+            }
         }
     }
 
@@ -1071,8 +1200,6 @@ _ibus_create_input_context (BusIBusImpl     *ibus,
         GCallback callback;
     } signals [] = {
         { "request-engine",      G_CALLBACK (_context_request_engine_cb) },
-        { "request-next-engine", G_CALLBACK (_context_request_next_engine_cb) },
-        { "request-prev-engine", G_CALLBACK (_context_request_prev_engine_cb) },
         { "engine-changed", G_CALLBACK (_context_engine_changed_cb) },
         { "focus-in",       G_CALLBACK (_context_focus_in_cb) },
         { "focus-out",      G_CALLBACK (_context_focus_out_cb) },
@@ -1090,6 +1217,10 @@ _ibus_create_input_context (BusIBusImpl     *ibus,
                           ibus);
     }
 
+    if (ibus->enable_by_default) {
+        bus_input_context_enable (context);
+    }
+
     path = ibus_service_get_path ((IBusService *) context);
     reply = ibus_message_new_method_return (message);
     ibus_message_append_args (reply,
@@ -1354,6 +1485,155 @@ _ibus_ping (BusIBusImpl     *ibus,
     return reply;
 }
 
+static IBusMessage *
+_ibus_get_use_sys_layout (BusIBusImpl     *ibus,
+                          IBusMessage     *message,
+                          BusConnection   *connection)
+{
+    IBusMessage *reply;
+
+    reply = ibus_message_new_method_return (message);
+    ibus_message_append_args (reply,
+            G_TYPE_BOOLEAN, &ibus->use_sys_layout,
+            G_TYPE_INVALID);
+
+    return reply;
+}
+
+static IBusMessage *
+_ibus_get_use_global_engine (BusIBusImpl     *ibus,
+                             IBusMessage     *message,
+                             BusConnection   *connection)
+{
+    IBusMessage *reply;
+
+    reply = ibus_message_new_method_return (message);
+    ibus_message_append_args (reply,
+            G_TYPE_BOOLEAN, &ibus->use_global_engine,
+            G_TYPE_INVALID);
+
+    return reply;
+}
+
+static IBusMessage *
+_ibus_get_global_engine (BusIBusImpl     *ibus,
+                         IBusMessage     *message,
+                         BusConnection   *connection)
+{
+    IBusMessage *reply;
+
+    if (ibus->use_global_engine && ibus->global_engine) {
+        IBusEngineDesc *desc = bus_engine_proxy_get_desc (ibus->global_engine);
+        if (desc != NULL) {
+            reply = ibus_message_new_method_return (message);
+            ibus_message_append_args (reply,
+                                      IBUS_TYPE_ENGINE_DESC, &desc,
+                                      G_TYPE_INVALID);
+            return reply;
+        }
+    }
+
+    reply = ibus_message_new_error (message, DBUS_ERROR_FAILED,
+                                    "No global engine.");
+    return reply;
+}
+
+static IBusMessage *
+_ibus_set_global_engine (BusIBusImpl     *ibus,
+                         IBusMessage     *message,
+                         BusConnection   *connection)
+{
+    gboolean retval;
+    IBusMessage *reply;
+    IBusError *error;
+    gchar *new_engine_name;
+    gchar *old_engine_name;
+
+    if (!ibus->use_global_engine) {
+        reply = ibus_message_new_error (message, DBUS_ERROR_FAILED,
+                                        "Global engine feature is disable.");
+        return reply;
+    }
+
+    retval = ibus_message_get_args (message,
+                                    &error,
+                                    G_TYPE_STRING, &new_engine_name,
+                                    G_TYPE_INVALID);
+    if (!retval) {
+        reply = ibus_message_new_error (message,
+                                        error->name,
+                                        error->message);
+        ibus_error_free (error);
+        return reply;
+    }
+
+    reply = ibus_message_new_method_return (message);
+    old_engine_name = NULL;
+
+    if (ibus->global_engine) {
+        old_engine_name = bus_engine_proxy_get_desc (ibus->global_engine)->name;
+    }
+
+    if (g_strcmp0 (new_engine_name, old_engine_name) == 0) {
+        /* If the user requested the same global engine, then we just enable the
+         * original one. */
+        if (ibus->focused_context) {
+            bus_input_context_enable (ibus->focused_context);
+        }
+        else if (ibus->global_engine) {
+            bus_engine_proxy_enable (ibus->global_engine);
+        }
+        return reply;
+    }
+
+    /* If there is a focused input context, then we just change the engine of
+     * the focused context, which will then change the global engine
+     * automatically. Otherwise, we need to change the global engine directly.
+     */
+    if (ibus->focused_context) {
+        _context_request_engine_cb (ibus->focused_context, new_engine_name, ibus);
+    }
+    else {
+        IBusEngineDesc *engine_desc = _find_engine_desc_by_name (ibus, new_engine_name);
+        if (engine_desc != NULL) {
+            BusEngineProxy *new_engine = bus_ibus_impl_create_engine (engine_desc);
+            if (new_engine != NULL) {
+                /* Enable the global engine by default, because the user
+                 * selected it explicitly. */
+                bus_engine_proxy_enable (new_engine);
+
+                /* Assume the ownership of the new global engine. Normally it's
+                 * done by the input context. But as we need to change the global
+                 * engine directly, so we need to do it here. */
+                g_object_ref_sink (new_engine);
+                bus_ibus_impl_set_global_engine (ibus, new_engine);
+
+                /* The global engine should already be referenced. */
+                g_object_unref (new_engine);
+            }
+        }
+    }
+
+    return reply;
+}
+
+static IBusMessage *
+_ibus_is_global_engine_enabled (BusIBusImpl     *ibus,
+                                IBusMessage     *message,
+                                BusConnection   *connection)
+{
+    IBusMessage *reply;
+    gboolean enabled = (ibus->use_global_engine && ibus->global_engine &&
+                        bus_engine_proxy_is_enabled (ibus->global_engine));
+
+    reply = ibus_message_new_method_return (message);
+    ibus_message_append_args (reply,
+                              G_TYPE_BOOLEAN, &enabled,
+                              G_TYPE_INVALID);
+    return reply;
+}
+
+
 static gboolean
 bus_ibus_impl_ibus_message (BusIBusImpl     *ibus,
                             BusConnection   *connection,
@@ -1383,6 +1663,11 @@ bus_ibus_impl_ibus_message (BusIBusImpl     *ibus,
         { IBUS_INTERFACE_IBUS, "ListActiveEngines",     _ibus_list_active_engines },
         { IBUS_INTERFACE_IBUS, "Exit",                  _ibus_exit },
         { IBUS_INTERFACE_IBUS, "Ping",                  _ibus_ping },
+        { IBUS_INTERFACE_IBUS, "GetUseSysLayout",       _ibus_get_use_sys_layout },
+        { IBUS_INTERFACE_IBUS, "GetUseGlobalEngine",    _ibus_get_use_global_engine },
+        { IBUS_INTERFACE_IBUS, "GetGlobalEngine",       _ibus_get_global_engine },
+        { IBUS_INTERFACE_IBUS, "SetGlobalEngine",       _ibus_set_global_engine },
+        { IBUS_INTERFACE_IBUS, "IsGlobalEngineEnabled", _ibus_is_global_engine_enabled },
     };
 
     ibus_message_set_sender (message, bus_connection_get_unique_name (connection));
@@ -1475,3 +1760,142 @@ bus_ibus_impl_registry_changed (BusIBusImpl *ibus)
     ibus_message_unref (message);
 
 }
+
+static void
+bus_ibus_impl_global_engine_changed (BusIBusImpl *ibus)
+{
+    g_assert (BUS_IS_IBUS_IMPL (ibus));
+
+    IBusMessage *message;
+
+    message = ibus_message_new_signal (IBUS_PATH_IBUS,
+                                       IBUS_INTERFACE_IBUS,
+                                       "GlobalEngineChanged");
+    ibus_message_append_args (message,
+                              G_TYPE_INVALID);
+    ibus_message_set_sender (message, IBUS_SERVICE_IBUS);
+
+    bus_dbus_impl_dispatch_message_by_rule (BUS_DEFAULT_DBUS, message, NULL);
+
+    ibus_message_unref (message);
+}
+
+gboolean
+bus_ibus_impl_filter_keyboard_shortcuts (BusIBusImpl     *ibus,
+                                         BusInputContext *context,
+                                         guint           keyval,
+                                         guint           modifiers,
+                                         guint           prev_keyval,
+                                         guint           prev_modifiers)
+{
+    static GQuark trigger = 0;
+    static GQuark next = 0;
+    static GQuark previous = 0;
+
+    if (trigger == 0) {
+        trigger = g_quark_from_static_string ("trigger");
+        next = g_quark_from_static_string ("next-engine-in-menu");
+        previous = g_quark_from_static_string ("previous-engine");
+    }
+
+    GQuark event = ibus_hotkey_profile_filter_key_event (ibus->hotkey_profile,
+                                                         keyval,
+                                                         modifiers,
+                                                         prev_keyval,
+                                                         prev_modifiers,
+                                                         0);
+
+    if (event == trigger) {
+        gboolean enabled = bus_input_context_is_enabled (context);
+        if (enabled) {
+            bus_input_context_disable (context);
+        }
+        else {
+            bus_input_context_enable (context);
+        }
+        return (enabled != bus_input_context_is_enabled (context));
+    }
+    if (event == next) {
+        if (bus_input_context_is_enabled(context)) {
+            bus_ibus_impl_context_request_next_engine_in_menu (ibus, context);
+        }
+        else {
+            bus_input_context_enable (context);
+        }
+        return TRUE;
+    }
+    if (event == previous) {
+        if (bus_input_context_is_enabled(context)) {
+            bus_ibus_impl_context_request_previous_engine (ibus, context);
+        }
+        else {
+            bus_input_context_enable (context);
+        }
+        return TRUE;
+    }
+    return FALSE;
+}
+
+static gchar*
+bus_ibus_impl_load_global_engine_name_from_config (BusIBusImpl *ibus)
+{
+    GValue value = { 0 };
+    gchar *global_engine_name = NULL;
+
+    g_assert (IBUS_IS_CONFIG (ibus->config));
+
+    if (ibus_config_get_value (ibus->config, "general", "global_engine", &value) &&
+        G_VALUE_TYPE (&value) == G_TYPE_STRING) {
+        global_engine_name = g_value_dup_string (&value);
+        g_value_unset (&value);
+    }
+
+    return global_engine_name;
+}
+
+static void
+bus_ibus_impl_save_global_engine_name_to_config (BusIBusImpl *ibus)
+{
+    g_assert (IBUS_IS_CONFIG (ibus->config));
+
+    if (ibus->use_global_engine && ibus->global_engine) {
+        GValue value = { 0 };
+        g_value_init (&value, G_TYPE_STRING);
+        g_value_set_static_string (&value, bus_engine_proxy_get_desc (ibus->global_engine)->name);
+
+        ibus_config_set_value (ibus->config, "general", "global_engine", &value);
+        g_value_unset (&value);
+    }
+}
+
+static gchar*
+bus_ibus_impl_load_global_previous_engine_name_from_config (BusIBusImpl *ibus)
+{
+    GValue value = { 0 };
+    gchar *global_previous_engine_name = NULL;
+
+    g_assert (IBUS_IS_CONFIG (ibus->config));
+
+    if (ibus_config_get_value (ibus->config, "general", "global_previous_engine", &value) &&
+        G_VALUE_TYPE (&value) == G_TYPE_STRING) {
+        global_previous_engine_name = g_value_dup_string (&value);
+        g_value_unset (&value);
+    }
+
+    return global_previous_engine_name;
+}
+
+static void
+bus_ibus_impl_save_global_previous_engine_name_to_config (BusIBusImpl *ibus)
+{
+    g_assert (IBUS_IS_CONFIG (ibus->config));
+
+    if (ibus->use_global_engine && ibus->global_previous_engine_name) {
+        GValue value = { 0 };
+        g_value_init (&value, G_TYPE_STRING);
+        g_value_set_static_string (&value, ibus->global_previous_engine_name);
+
+        ibus_config_set_value (ibus->config, "general", "global_previous_engine", &value);
+        g_value_unset (&value);
+    }
+}
