@@ -24,6 +24,7 @@
 #include <unistd.h>
 #include <glib/gstdio.h>
 #include <gio/gio.h>
+#include <dbus/dbus.h>
 #include "ibusbus.h"
 #include "ibusinternal.h"
 #include "ibusshare.h"
@@ -38,6 +39,7 @@
 enum {
     CONNECTED,
     DISCONNECTED,
+    GLOBAL_ENGINE_CHANGED,
     LAST_SIGNAL,
 };
 
@@ -54,7 +56,13 @@ typedef struct _IBusBusPrivate IBusBusPrivate;
 
 static guint    bus_signals[LAST_SIGNAL] = { 0 };
 
+static IBusBus *_bus = NULL;
+
 /* functions prototype */
+static GObject* ibus_bus_constructor    (GType          type,
+                                         guint          n_params,
+                                         GObjectConstructParam
+                                                        *params);
 static void     ibus_bus_destroy        (IBusObject     *object);
 static void     ibus_bus_watch_dbus_signal
                                         (IBusBus        *bus);
@@ -74,10 +82,12 @@ ibus_bus_new (void)
 static void
 ibus_bus_class_init (IBusBusClass *klass)
 {
+    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
     IBusObjectClass *ibus_object_class = IBUS_OBJECT_CLASS (klass);
 
     g_type_class_add_private (klass, sizeof (IBusBusPrivate));
 
+    gobject_class->constructor = ibus_bus_constructor;
     ibus_object_class->destroy = ibus_bus_destroy;
 
     // install signals
@@ -114,6 +124,39 @@ ibus_bus_class_init (IBusBusClass *klass)
             ibus_marshal_VOID__VOID,
             G_TYPE_NONE,
             0);
+
+    /**
+     * IBusBus::global-engine-changed:
+     *
+     * Emitted when global engine is changed.
+     *
+     * <note><para>Argument @user_data is ignored in this function.</para></note>
+     */
+    bus_signals[GLOBAL_ENGINE_CHANGED] =
+        g_signal_new (I_("global-engine-changed"),
+            G_TYPE_FROM_CLASS (klass),
+            G_SIGNAL_RUN_LAST,
+            0,
+            NULL, NULL,
+            ibus_marshal_VOID__VOID,
+            G_TYPE_NONE,
+            0);
+}
+
+static gboolean
+_connection_ibus_signal_cb (IBusConnection *connection,
+                            IBusMessage    *message,
+                            IBusBus        *bus)
+{
+    IBusBusPrivate *priv;
+    priv = IBUS_BUS_GET_PRIVATE (bus);
+
+    if (ibus_message_is_signal (message, IBUS_INTERFACE_IBUS,
+                                "GlobalEngineChanged")) {
+        g_signal_emit (bus, bus_signals[GLOBAL_ENGINE_CHANGED], 0);
+        return TRUE;
+    }
+    return FALSE;
 }
 
 static void
@@ -130,6 +173,9 @@ _connection_destroy_cb (IBusConnection  *connection,
     g_signal_handlers_disconnect_by_func (priv->connection,
                                           G_CALLBACK (_connection_destroy_cb),
                                           bus);
+    g_signal_handlers_disconnect_by_func (priv->connection,
+                                          G_CALLBACK (_connection_ibus_signal_cb),
+                                          bus);
     g_object_unref (priv->connection);
     priv->connection = NULL;
 
@@ -167,6 +213,7 @@ ibus_bus_connect (IBusBus *bus)
 
     if (priv->connection) {
         ibus_bus_hello (bus);
+
         g_signal_connect (priv->connection,
                           "destroy",
                           (GCallback) _connection_destroy_cb,
@@ -176,6 +223,18 @@ ibus_bus_connect (IBusBus *bus)
         if (priv->watch_dbus_signal) {
             ibus_bus_watch_dbus_signal (bus);
         }
+
+        /** Watch ibus signals. */
+        const gchar *rule =
+            "type='signal',"
+            "path='" IBUS_PATH_IBUS "',"
+            "interface='" IBUS_INTERFACE_IBUS "'";
+
+        ibus_bus_add_match (bus, rule);
+        g_signal_connect (priv->connection,
+                          "ibus-signal",
+                          (GCallback) _connection_ibus_signal_cb,
+                          bus);
     }
 }
 
@@ -193,7 +252,7 @@ _changed_cb (GFileMonitor       *monitor,
 
     if (ibus_bus_is_connected (bus))
         return;
-    
+
     ibus_bus_connect (bus);
 }
 
@@ -236,6 +295,27 @@ ibus_bus_init (IBusBus *bus)
     g_free (path);
 }
 
+static GObject*
+ibus_bus_constructor (GType                  type,
+                      guint                  n_params,
+                      GObjectConstructParam *params)
+{
+    GObject *object;
+
+    /* share one IBusBus instance in whole application */
+    if (_bus == NULL) {
+        object = G_OBJECT_CLASS (ibus_bus_parent_class)->constructor (type, n_params, params);
+        /* make bus object sink */
+        g_object_ref_sink (object);
+        _bus = IBUS_BUS (object);
+    }
+    else {
+        object = g_object_ref (_bus);
+    }
+
+    return object;
+}
+
 static void
 ibus_bus_destroy (IBusObject *object)
 {
@@ -245,6 +325,9 @@ ibus_bus_destroy (IBusObject *object)
     bus = IBUS_BUS (object);
     priv = IBUS_BUS_GET_PRIVATE (bus);
 
+    g_assert (_bus == (IBusBus *)object);
+    _bus = NULL;
+
     if (priv->monitor) {
         g_object_unref (priv->monitor);
         priv->monitor = NULL;
@@ -706,7 +789,7 @@ ibus_bus_get_name_owner (IBusBus        *bus,
                                       DBUS_PATH_DBUS,
                                       DBUS_INTERFACE_DBUS,
                                       "GetNameOwner",
-                                      G_TYPE_STRING, &owner,
+                                      G_TYPE_STRING, &name,
                                       G_TYPE_INVALID);
 
     if (reply) {
@@ -885,3 +968,133 @@ ibus_bus_get_config (IBusBus *bus)
 
     return priv->config;
 }
+
+gboolean
+ibus_bus_get_use_sys_layout (IBusBus *bus)
+{
+    g_assert (IBUS_IS_BUS (bus));
+
+    IBusMessage *reply = NULL;
+    IBusError *error = NULL;
+    gboolean use_sys_layout = FALSE;
+
+    reply = ibus_bus_call_with_reply (bus,
+                                      IBUS_SERVICE_IBUS,
+                                      IBUS_PATH_IBUS,
+                                      IBUS_INTERFACE_IBUS,
+                                      "GetUseSysLayout",
+                                      G_TYPE_INVALID);
+    if (reply) {
+        if (!ibus_message_get_args (reply, &error, G_TYPE_BOOLEAN,
+                                    &use_sys_layout, G_TYPE_INVALID)) {
+            g_warning ("%s: %s", error->name, error->message);
+            ibus_error_free (error);
+        }
+
+        ibus_message_unref (reply);
+    }
+
+    return use_sys_layout;
+}
+
+gboolean
+ibus_bus_get_use_global_engine (IBusBus *bus)
+{
+    g_assert (IBUS_IS_BUS (bus));
+
+    IBusMessage *reply = NULL;
+    IBusError *error = NULL;
+    gboolean use_global_engine = FALSE;
+
+    reply = ibus_bus_call_with_reply (bus,
+                                      IBUS_SERVICE_IBUS,
+                                      IBUS_PATH_IBUS,
+                                      IBUS_INTERFACE_IBUS,
+                                      "GetUseGlobalEngine",
+                                      G_TYPE_INVALID);
+    if (reply) {
+        if (!ibus_message_get_args (reply, &error, G_TYPE_BOOLEAN,
+                                    &use_global_engine, G_TYPE_INVALID)) {
+            g_warning ("%s: %s", error->name, error->message);
+            ibus_error_free (error);
+        }
+
+        ibus_message_unref (reply);
+    }
+
+    return use_global_engine;
+}
+
+gboolean
+ibus_bus_is_global_engine_enabled (IBusBus *bus)
+{
+    g_assert (IBUS_IS_BUS (bus));
+
+    IBusMessage *reply = NULL;
+    IBusError *error = NULL;
+    gboolean global_engine_enabled = FALSE;
+
+    reply = ibus_bus_call_with_reply (bus,
+                                      IBUS_SERVICE_IBUS,
+                                      IBUS_PATH_IBUS,
+                                      IBUS_INTERFACE_IBUS,
+                                      "IsGlobalEngineEnabled",
+                                      G_TYPE_INVALID);
+    if (reply) {
+        if (!ibus_message_get_args (reply, &error, G_TYPE_BOOLEAN,
+                                    &global_engine_enabled, G_TYPE_INVALID)) {
+            g_warning ("%s: %s", error->name, error->message);
+            ibus_error_free (error);
+        }
+
+        ibus_message_unref (reply);
+    }
+
+    return global_engine_enabled;
+}
+
+IBusEngineDesc *
+ibus_bus_get_global_engine (IBusBus *bus)
+{
+    g_assert (IBUS_IS_BUS (bus));
+
+    IBusMessage *reply = NULL;
+    IBusError *error = NULL;
+    IBusEngineDesc *global_engine = NULL;
+
+    reply = ibus_bus_call_with_reply (bus,
+                                      IBUS_SERVICE_IBUS,
+                                      IBUS_PATH_IBUS,
+                                      IBUS_INTERFACE_IBUS,
+                                      "GetGlobalEngine",
+                                      G_TYPE_INVALID);
+    if (reply) {
+        if (!ibus_message_get_args (reply, &error, IBUS_TYPE_ENGINE_DESC,
+                                    &global_engine, G_TYPE_INVALID)) {
+            g_warning ("%s: %s", error->name, error->message);
+            ibus_error_free (error);
+        }
+
+        ibus_message_unref (reply);
+    }
+
+    return global_engine;
+}
+
+gboolean
+ibus_bus_set_global_engine (IBusBus     *bus,
+                            const gchar *global_engine)
+{
+    g_assert (IBUS_IS_BUS (bus));
+
+    gboolean result;
+    result = ibus_bus_call (bus,
+                            IBUS_SERVICE_IBUS,
+                            IBUS_PATH_IBUS,
+                            IBUS_INTERFACE_IBUS,
+                            "SetGlobalEngine",
+                            G_TYPE_STRING, &global_engine,
+                            G_TYPE_INVALID);
+
+    return result;
+}
