@@ -19,6 +19,7 @@
  * Boston, MA 02111-1307, USA.
  */
 #include <stdarg.h>
+#include <dbus/dbus.h>
 #include "ibusmainloop.h"
 #include "ibusmessage.h"
 #include "ibusconnection.h"
@@ -737,7 +738,7 @@ ibus_connection_call_with_reply_valist (IBusConnection     *connection,
     g_assert (path != NULL);
     g_assert (interface != NULL);
     g_assert (member != NULL);
-    g_return_val_if_fail (ibus_connection_is_connected (connection), FALSE);
+    g_assert (ibus_connection_is_connected (connection));
 
     IBusConnectionPrivate *priv;
     priv = IBUS_CONNECTION_GET_PRIVATE (connection);
